from ..tools.ones import ones
from ..programs.move import move

# add
# ===
# dst     : positive integer, the index of the destination register which also
#         :   must contain the first addend
# src     : positive integer, the index of the source register which must
#         :   contain the second addend
# tmp     : positive integer, the index of the temporary register
# returns : string, the 1# program which will calculate the sum, in backward
#         :   binary, of the contents of registers ``dst'' and ``src'' and store
#         :   the result back into register ``dst'' while using register ``tmp''
#         :   as temporary scratch space
def add(dst, src, tmp):
  p = ones(dst)+'#####'  # Cases on register ``dst''
  p += '111###'          # Go forward three
  p += '11111 1###'      # Go forward six
  p += '11111 1111###'   # Go forward nine
  p += ones(src)+'#####' # Cases on register ``src''
                         # Go forward thirty-five
  p += '11111 11111 11111 11111 11111 11111 11111###'
                         # Go forward twenty-six
  p += '11111 11111 11111 11111 11111 1###'
                         # Go forward twenty-seven
  p += '11111 11111 11111 11111 11111 11###'
  p += ones(src)+'#####' # Cases on register ``src''
                         # Go forward twenty-three
  p += '11111 11111 11111 11111 111###'
                         # Go forward twenty-eight
  p += '11111 11111 11111 11111 11111 111###'
                         # Go forward twenty-one
  p += '11111 11111 11111 11111 1###'
  p += ones(src)+'#####' # Cases on register ``src''
                         # Go forward twenty-one
  p += '11111 11111 11111 11111 1###'
                         # Go forward eighteen
  p += '11111 11111 11111 111###'
                         # Go forward nineteen
  p += '11111 11111 11111 1111###'
  p += ones(dst)+'#####' # Cases on register ``dst''
  p += '111###'          # Go forward three
  p += '11111 1###'      # Go forward six
  p += '11111 1111###'   # Go forward nine
  p += ones(src)+'#####' # Cases on register ``src''
                         # Go forward eleven
  p += '11111 11111 1###'
                         # Go forward sixteen
  p += '11111 11111 11111 1###'
  p += '11111 1111###'   # Go forward nine
  p += ones(src)+'#####' # Cases on register ``src''
                         # Go forward thirteen
  p += '11111 11111 111###'
  p += '11111 11111###'  # Go forward ten
                         # Go forward eleven
  p += '11111 11111 1###'
  p += ones(src)+'#####' # Cases on register ``src''
  p += '111###'          # Go forward three
  p += '11111 111###'    # Go forward eight
  p += '1###'            # Go forward one; no op
  p += ones(tmp)+'#'     # Append ``1'' to register ``tmp''
                         # Go backward thirty-three
  p += '11111 11111 11111 11111 11111 11111 111####'
  p += ones(tmp)+'##'    # Append ``#'' to register ``tmp''
                         # Go backward thirty-five
  p += '11111 11111 11111 11111 11111 11111 11111####'
  p += ones(tmp)+'#'     # Append ``1'' to register ``tmp''
                         # Go backward twenty-one
  p += '11111 11111 11111 11111 1####'
  p += ones(tmp)+'##'    # Append ``#'' to register ``tmp''
                         # Go backward twenty-three
  p += '11111 11111 11111 11111 111####'
  p += move(tmp, dst)    # Move the contents of register ``tmp'' to register
                         #   ``dst''
  return p



from ..tools.instruction_set import enqueue_one as eo
from ..tools.instruction_set import jump_forward as jf
from ..tools.instruction_set import jump_backward as jb
from ..tools.instruction_set import dequeue_and_branch as db
from ..tools.labels_to_offsets import labels_to_offsets
from ..programs.clear import clear

# compare
# =======
# a       : positive integer, the index of the register which contains the first
#         :   string in the comparison and in which the symbol encoding the
#         :   result of the comparison will be written
# b       : positive integer, the index of the register which contains the
#         :   second string in the comparison
# returns : string, the 1# program which will compare the strings in registers
#         :   ``a'' and ``b,'' decide if they are symbol-for-symbol the same,
#         :   and if they are the same write a ``1'' back into register ``a''
#         :  but otherwise clear register ``a'' (if it is not already clear);
#         :  register``b'' will be empty when the computation is finished
def compare(a, b):
  p =  '<continue>'
  p +=   db(a)
  p +=   jf('<a_empty>')
  p +=   jf('<a_one>')
  p +=   jf('<a_sharp>')
  p +=   '<a_empty>'
  p +=     db(b)
  p +=     jf('<a_empty_b_empty>')
  p +=     jf('<a_empty_b_one>')
  p +=     jf('<a_empty_b_sharp>')
  p +=     '<a_empty_b_empty>'
             # a *does* equal b
  p +=       eo(a)
  p +=       jf('<break>')
  p +=     '<a_empty_b_one>'
             # a does *not* equal 
  p +=       clear(b)
  p +=       jf('<break>')
  p +=     '<a_empty_b_sharp>'
             # a does *not* equal b
  p +=       clear(b)
  p +=       jf('<break>')
  p +=   '<a_one>'
  p +=      db(b)
  p +=      jf('<a_one_b_empty>')
  p +=      jf('<a_one_b_one>')
  p +=      jf('<a_one_b_sharp>')
  p +=      '<a_one_b_empty>'
              # a does *not* equal b
  p +=        clear(a)
  p +=        jf('<break>')
  p +=      '<a_one_b_one>'
              # a equals b *so far*
  p +=        jb('<continue>')
  p +=      '<a_one_b_sharp>'
              # a does *not* equal b
  p +=        clear(a)
  p +=        clear(b)
  p +=        jf('<break>')
  p +=   '<a_sharp>'
  p +=     db(b)
  p +=     jf('<a_sharp_b_empty>')
  p +=     jf('<a_sharp_b_one>')
  p +=     jf('<a_sharp_b_sharp>')
  p +=     '<a_sharp_b_empty>'
             # a does *not* equal b
  p +=       clear(a)
  p +=       jf('<break>')
  p +=     '<a_sharp_b_one>'
             # a does *not* equal b
  p +=       clear(a)
  p +=       clear(b)
  p +=       jf('<break>')
  p +=     '<a_sharp_b_sharp>'
             # a equals b *so far*
  p +=       jb('<continue>')
  p += '<break>'
  return labels_to_offsets(p)



from ..tools.instruction_set import enqueue_one as eo
from ..tools.instruction_set import enqueue_sharp as es
from ..tools.instruction_set import jump_forward as jf
from ..tools.instruction_set import jump_backward as jb
from ..tools.instruction_set import dequeue_and_branch as db
from ..tools.labels_to_offsets import labels_to_offsets
from ..programs.clear import clear
from ..programs.move import move
from ..programs.copy import copy
from ..programs.compare import compare

# div
# =======
# dividend      : positive integer, the index of the register in which the
#               :   dividend is passed in unary and to which the result is
#               :   written at the end of the computation (default 1)
# divisor       : positive integer, the index of the register in which the
#               :   divisor is passed in unary; on a successful return of the
#               :   1# div program, this register will be empty (default 2)
# divisor_tmp   : positive integer, the index of the temporary divisor register
#               :   that allows us to preserve the divisor across destructive
#               :   register-compare operations; this register must be passed
#               :   empty and, on a successful return of the 1# div program,
#               :   it will remain empty (default 3)
# quotient      : positive integer, the index of the register in which we build
#               :   up the quotient during the division computation; this
#               :   register must be passed empty and, on a successful return
#               :   of the 1# div program, this register will remain empty
#               :   (default 4)
# remainder     : positive integer, the index of the counter register that we
#               :   use to repeatedly count up to the divisor during the divison
#               :   operation; this register must be passed in empty and, on a
#               :   successful return of the 1# div program, this register will
#               :   remain empty (default 5)
# remainder_tmp : positive integer, the index of the temporary remainder
#               :   register that allows us to preserve the current remainder
#               :   across destructive register-compare operations; this
#               :   this register must be passed empty and, on a succesful
#               :   return of the 1# div program, it will remain empty
#               :   (default 6)
# copy_tmp      : positive integer, the index of the register that is used
#               :   as the temporary in for copy operations; it should be
#               :   passed empty and the 1# program will return with it empty
#               :   (default 7)
# error         : positive integer, the index of the error register; this
#               :   register should be passed empty and it will be returned
#               :   containing a one in some error events, but very few
#               :   potential errors are actually caught; for example, there is
#               :   not logic present to check for a zero divisor (default 8)
#               :
# returns       : string, the 1# program which divides the contents of register
#               :   ``dividend'' by register ``divisor'' and writes the result
#               :   back into register ``dividend''
#
#   examples
#   ========
#     >>> onesharp(div(1,2,3,4,5,6,7,8), ['111','11'])
#     '1#1'
#     >>> onesharp(div(1,2,3,4,5,6,7,8), ['111111','111'])
#     '11#'
#
def div(
  dividend     =1,
  divisor      =2,
  divisor_tmp  =3,
  quotient     =4,
  remainder    =5,
  remainder_tmp=6,
  copy_tmp     =7,
  error        =8
):
  p =  '<continue>'
  p +=   db(dividend)
  p +=   jf('<dividend_empty>')
  p +=   jf('<dividend_one>')
  p +=   jf('<dividend_sharp>')
  p +=   '<dividend_empty>'
           # Division complete
  p +=     move(quotient, dividend)
  p +=     es(dividend)
  p +=     move(remainder, dividend)
  p +=     clear(divisor)
  p +=     jf('<break>')
  p +=   '<dividend_one>'
  p +=     eo(remainder)
  p +=     copy(divisor, divisor_tmp, copy_tmp)
  p +=     copy(remainder, remainder_tmp, copy_tmp)
  p +=     compare(divisor_tmp, remainder_tmp)
  p +=     db(divisor_tmp)
  p +=     jf('<not_equal>')
  p +=     jf('<yes_equal>')
  p +=     jf('<error>') # Compare should never return a sharp
  p +=     '<not_equal>'
  p +=       jb('<continue>')
  p +=     '<yes_equal>'
  p +=       clear(remainder)
  p +=       eo(quotient)
  p +=       jb('<continue>')
  p +=   '<dividend_sharp>'
           # Error since we expect the dividend to be in unary
  p +=     jf('<error>')
  p += '<error>'
  p +=   eo(error)
  p += '<break>'
  return labels_to_offsets(p)



from ..tools.ones import ones

# clear
# =====
# reg     : positive integer, the index of the register to clear
# returns : string, the 1# program to clear register ``reg''
def clear(reg):
  p = ones(reg)+'#####' # Cases on register ``reg''
  p += '111###'         # Go forward three
  p += '11####'         # Go backward two
  p += '111####'        # Go backward three
  return p



from ..tools.ones import ones
from ..programs.move import move

# successor
# =========
# reg     : positive integer, the index of the register for whose contents
#         :   the successor will be calculated
# tmp     : positive integer, the index of the temporary register
# returns : string, the 1# program which will find the successor of the
#         :   (nonnegative) backward binary integer in register ``reg''
#         :   and place the successor back into register ``reg'' while using
#         :   register ``tmp'' as temporary scratch space
def successor(reg, tmp):
  p = ones(reg)+'#####'      # Cases on register ``reg''
  p += '11111 11111 111###'  # Go forward thirteen
  p += '11111 11111###'      # Go forward ten 
  p += ones(tmp)+'#'         # Append ``1'' to register ``tmp''
  p += move(reg, tmp)        # Move the remaining contents of register ``reg''
                             #   to register ``tmp''; seven instructions
  p += '1111###'             # Go forward four
  p += ones(tmp)+'##'        # Append ``#'' to register ``tmp''
  p += '11111 11111 111####' # Go backward thirteen
  p += ones(tmp)+'#'         # Append ``1'' to register ``tmp''
  p += move(tmp, reg)        # Move the contents of register ``tmp'' to
                             #   register ``reg''
  return p



from ..tools.ones import ones

# move
# ====
# src     : positive integer, the index of the source register
# dst     : positive integer, the index of the destination register
# returns : string, the 1# program to move the contents of register ``src'' to
#         :   register ``dst''
def move(src, dst):
  p = ones(src)+'#####' # Cases on register ``src''
  p += '11111 1###'     # Go forward six
  p += '111###'         # Go forward three
  p += ones(dst)+'##'   # Append ``#'' to register ``dst''
  p += '1111####'       # Go backward four
  p += ones(dst)+'#'    # Append ``1'' to register ``dst''
  p += '11111 1####'    # Go backward six
  return p



from ..tools.instruction_set import enqueue_one as eo
from ..tools.instruction_set import enqueue_sharp as es
from ..tools.instruction_set import jump_forward as jf
from ..tools.instruction_set import jump_backward as jb
from ..tools.instruction_set import dequeue_and_branch as db
from ..tools.labels_to_offsets import labels_to_offsets
from ..programs.move import move

# write
# =====
# reg     : positive integer, the index of the register on whose contents the
#         :   the write program will be executed and into which the output
#         :   program-text will be written (default is 1)
# tmp     : positive integer, the index of the temporary register (default is 2)
# returns : string, the 1# program with the property that when it is run with
#         :   contents ``x'' in register ``reg'' and register ``tmp'' empty,
#         :   we eventually halt with contents ``y'' in regsiter ``reg'' and
#         :   register ``tmp'' empty, where ``y'' is a program in the 1#
#         :   language such that when ``y'' is executed (with no arguments),
#         :   the result is to write ``x'' into register ``reg''
def write(reg=1, tmp=2):
  p =  '<top>'
  p +=   db(reg)
  p +=   jf('<empty>')
  p +=   jf('<one>')
  p +=   '<sharp>'
  for _ in range(reg):
    p +=   eo(tmp)
  p +=     es(tmp)
  p +=     es(tmp)
  p +=     jb('<top>')
  p +=   '<one>'
  for _ in range(reg):
    p +=   eo(tmp)
  p +=     es(tmp)
  p +=     jb('<top>')
  p +=   '<empty>'
  p +=     move(tmp, reg)
  return labels_to_offsets(p)



from ..tools.instruction_set import enqueue_one as eo
from ..tools.instruction_set import enqueue_sharp as es
from ..tools.instruction_set import jump_forward as jf
from ..tools.instruction_set import jump_backward as jb
from ..tools.instruction_set import dequeue_and_branch as db
from ..tools.instruction_set import dequeue as dq
from ..tools.labels_to_offsets import labels_to_offsets
from ..programs.clear import clear
from ..programs.move import move
from ..programs.copy import copy

# The registers used in the universal 1# program
# ==============================================
PROG_TXT         =  1 # Program text : the index of the register that contains
                      #   the text of the 1# program that we are running the
                      #   universal program on
PROG_TXT_TMP     =  2 # Program text temporary : the index of the temporary
                      #   register used when cycling through the program text
INSTR_IDX        =  3 # Instruction index : the index of the register that
                      #   contains the index of the instruction in the
                      #   the program text which we would like to execute or
                      #   which is in the process of being executed; this value,
                      #   in addition to being call the ``instruction index'',
                      #   can synonymously be called the ``program counter'' and
                      #   is encoded in unary, with an empty register meaning
                      #   zero (i.e., the first instruction in the program),
                      #   ``1'' meaning one, ``11'' meaning two, etc.)
INSTR_IDX_TMP    =  4 # Instruction index temporary : the index of the register
                      #   that is used to save the instruction index when we
                      #   are counting down on the actual instruction index
                      #   register to locate an instruction in the program text
INSTR_ONES       =  5 # Instruction ones : the index of the instruction ones
                      #   register; the process of fetching an instruction
                      #   writes the number of ones in that instruction into
                      #   the instruction ones register; e.g., after the
                      #   instruction ``1111###'' is fetched, the contents of
                      #   this register will be ``1111''
INSTR_ONES_TMP   =  6 # Instruction ones temporary : the index of the register
                      #   used to save the instruction ones register when we
                      #   are doing operations directly on that register
INSTR_SHARPS     =  7 # Instruction sharps : the index of the instruction sharps
                      #   register; the process of fetching an instruction
                      #   writes the number of sharps in that instruction into
                      #   the instruction sharps register; e.g., after the
                      #   instruction ``1111###'' is fetched, the contents of
                      #   this register will be ``###''
INSTR_ONE_SAVE   =  8 # Instruction one save : the index of the temporary
                      #   register that is used when probing the boundary
                      #   between instructions in the fetch instruction module
REG_FILE         =  9 # Register file : the index of the register into which all
                      #   the registers of the program we are executing are
                      #   encoded via the encoding defined in the TRM web
                      #   text by L. Moss
REG_FILE_TMP     = 10 # Register file temporary : the index of the temporary
                      #   register used when cycling through the contents of the
                      #   register file
ENQUEUE_VAL_REG  = 11 # Enqueue value regsiter : the index of the register into
                      #   which the value to enqueue is written before control
                      #   is transfered to the enqueue module
VAL_DEQUEUED_REG = 12 # Value dequeued register : the index of the register into
                      #   which the value dequeued from the register file is
                      #   written before a call to the dequeue module returns
COPY_TMP         = 13 # Copy temporary : the index of the temporary register
                      #   that used as scratch space during copy operations 
EXCEPTION_REG    = 14 # Exception register : the index of the exception
                      #   register; there is not much error handling
                      #   (i.e., extra code to aid debugging)
                      #   in this program but, where it is present, if an
                      #   exception occurs a 1 will be written into the
                      #   exception register

def _fetch_instruction(
  prog_txt,
  prog_txt_tmp,
  instr_idx,
  instr_idx_tmp,
  instr_ones,
  instr_sharps,
  instr_one_save,
  copy_tmp
):
  p =  '<fetch-instruction>'
  p +=   '<get-ones-loop>'
  p +=     db(prog_txt)
  p +=     jf('<epilogue>') # jf('<end>')
  p +=     jf('<handle-a-one>')
  p +=     jf('<get-sharps-prologue>')
  p +=     '<handle-a-one>'
  p +=       eo(instr_ones)
  p +=       jb('<get-ones-loop>')
  p +=   '<get-sharps-prologue>'
  p +=     es(instr_sharps)
  p +=     '<get-sharps-loop>'
  p +=       db(prog_txt)
  p +=       jf('<tests>')
  p +=       jf('<between-instructions>')
  p +=       jf('<handle-a-sharp>')
  p +=       '<handle-a-sharp>'
  p +=         es(instr_sharps)
  p +=         jb('<get-sharps-loop>')
  p +=   '<between-instructions>'
  p +=     eo(instr_one_save)
  p +=     '<tests>'
  p +=       db(instr_idx)
  p +=       jf('<epilogue>')
  p +=       jf('<reloop>')
  p +=       jf('<error>')
  p +=       '<reloop>'
  p +=         move(instr_ones, prog_txt_tmp)
  p +=         move(instr_one_save, instr_ones)
  p +=         move(instr_sharps, prog_txt_tmp)
  p +=         eo(instr_idx_tmp)
  p +=         jb('<get-ones-loop>')
  p +=   '<epilogue>'
  p +=     copy(instr_ones, prog_txt_tmp, copy_tmp)
  p +=     copy(instr_sharps, prog_txt_tmp, copy_tmp)
  p +=     move(instr_one_save, prog_txt_tmp)
  p +=     move(prog_txt, prog_txt_tmp)
  p +=     move(prog_txt_tmp, prog_txt)
  p +=     move(instr_idx_tmp, instr_idx)
  p +=   '<end><error>'
  return labels_to_offsets(p)

def _populate_register_file(
  reg_file,
  reg_file_tmp,
  up_to,
  up_to_tmp,
  copy_tmp
):
  p =  copy(up_to, up_to_tmp, copy_tmp)
  p += '<populate-register-file>'
  p +=   db(up_to_tmp)
  p +=   jf('<epilogue>')
  p +=   jf('<check-reg-init-loop>')
  p +=   jf('<error>')
  p +=   '<check-reg-init-loop>'
  p +=     db(reg_file)
  p +=     jf('<handle-empty>')
  p +=     jf('<handle-one>')
  p +=     jf('<handle-sharp>')
  p +=     '<handle-one>'
  p +=       eo(reg_file_tmp)
  p +=       db(reg_file)
  p +=       jf('<error>')
  p +=       jf('<handle-one-one>')
  p +=       jf('<handle-one-sharp>')
  p +=       '<handle-one-one>'
  p +=         eo(reg_file_tmp)
  p +=         jb('<check-reg-init-loop>')
  p +=       '<handle-one-sharp>'
  p +=         es(reg_file_tmp)
  p +=         jb('<check-reg-init-loop>')
  p +=     '<handle-sharp>'
  p +=       es(reg_file_tmp)
  p +=       db(reg_file)
  p +=       jf('<error>')
  p +=       jf('<error>')
  p +=         es(reg_file_tmp)
  p +=         jb('<populate-register-file>')
  p +=     '<handle-empty>'
  p +=       es(reg_file_tmp)
  p +=       es(reg_file_tmp)
  p +=       jb('<populate-register-file>')
  p +=   '<epilogue>'
  p +=     move(reg_file, reg_file_tmp)
  p +=     move(reg_file_tmp, reg_file)
  p +=   '<error>'
  return labels_to_offsets(p)

# _enqueue_or_dequeue
# ===================
#                :
# reg_file       : positive integer, the index of the register in which all the
#                :   registers of the program we are executing via the
#                :   universal program are encoded; we expect the registers
#                :   to be encoded using the scheme defined in the universal
#                :   program chapter of the TRM web text by L. Moss
#                :
# reg_file_tmp   : positive integer, the index of the temporary register used
#                :   when cycling through the register-file register
#                :
# reg_to_get     : positive integer, the index of the register that we are
#                :   either enqueuing a character into or dequeuing a character
#                :   from
#                :
# reg_to_get_tmp : positive integer, the index of the register that we use to
#                :   to save the contents of the ``reg_to_get'' register while
#                :   we are doing processing directly on that register
#                :
# copy_tmp       : positive integer, the index of the register that we use for
#                :   temporary scratch space during copy operations
#                :
# is_enqueue     : boolean, whether to return the 1# code module for enqueue
#                :   or dequeue; if ``True'' then enqueue, and if ``false''
#                :   then dequeue; this is the only variable in this entire file
#                :   (universal.py) which is a sort of ``meta'' variable with
#                :   respect to the 1# program that is being generated and which
#                :   isn't included in the 1# program itself; all other
#                :   variables in this file are positive integers which each
#                :   refer to the index of a Moss machine register that is
#                :   operated on in the actual 1# program that is generated
#                :   by the Python code in this file
#                :
# enqueue_val    : positive integer, the index of the register into which the
#                :   the value is passed to this module to enqueue into the
#                :   register file in the case of an enqueue operation; in the
#                :   case of a dequeue operation, this register will not be used
#                :
# val_dequeued   : positive integer, the index of the register into which the
#                :   the value will be stored that is dequeued from the register
#                :   file in the case of a dequeue operation; in the case of an
#                :   enqueue operation, this register will not be used
#                :
# returns        : string, the 1# subprogram of the universal program which will
#                :   either enqueue a character into a specified register in the
#                :   the register file or will dequeue a character from a
#                :   specified register in the register file, depending on the
#                :   value specified for the ``is_enqueue'' argument
#                :
#   Summary of the 1# code modules that are returned for enqueuing and dequeuing
#   ----------------------------------------------------------------------------
#           The enqueue and dequeue modules both begin with a call to the
#   ``_populate_regsiter_file'' module for which the index of the register that
#   will be enqueued to or dequeued from is passed. The result of this call is
#   that the register file is initialized up to the index of the register that
#   will be enqueued to or dequeued from, if it is not already. For example,
#   if the register file contains ``11 ## ##'', encoding an R1 contents of
#   ``1'', an empty R2, and the remaining registers uninitialized, and if
#   further the enqueue or dequeue operation that we are preparing to perform
#   will operate on R4, then our call to the populate-register-file module will
#   result in R3 and R4 being initialized in the regsiter file so the new
#   contents are ``11 ## ## ## ##''. Initializing the register in the register
#   file that we will operate on before we begin the operation simplifies coding
#   the operation itself in 1#.
#           After the register file has been initialized up to the register we
#   will operate on, we next cycle through the register file until we reach
#   the index of that register to operate on. Once the register has been
#   reached, the action is of course different for an enqueue versus a dequeue
#   operation. For enqueue, we continue to cycle through the current register
#   until we reach the end-of-register code ``##''. At this point we append the
#   value that we are enqueuing onto the cycle-temporary-register
#   ``reg_file_tmp'', then we copy the remaining contents of the register file
#   onto the cycle register, followed by copying the cycle register back into
#   the ``reg_file'' register. At this point the enqueue operation is complete.
#           In the case of a dequeue operation, the value encoded into the
#   register of interest in the first position is read, without copying it into
#   the cycle register. Instead it is copied into the ``val_dequeued'' register
#   to be returned to the calling module. After this operation is complete,
#   the remaining contents, if any, of the register that we just dequeued from
#   are copied onto the cycle register, and after that all of the remaining
#   contents of the register file are copied onto the cycle register, and
#   finally the contents of the cycle register are moved back into the register
#   file, to complete the cycle. At this point the dequeue operation is
#   complete.
#
def _enqueue_or_dequeue(
  reg_file,
  reg_file_tmp,
  reg_to_get,
  reg_to_get_tmp,
  copy_tmp,
  is_enqueue,
  enqueue_val,
  val_dequeued
):
  assert type(reg_file) is int and reg_file > 0
  assert type(reg_file_tmp) is int and reg_file_tmp > 0
  assert type(reg_to_get) is int and reg_to_get > 0
  assert type(reg_to_get_tmp) is int and reg_to_get_tmp > 0
  assert type(copy_tmp) is int and copy_tmp > 0
  assert type(is_enqueue) is bool and is_enqueue in [True,False]
  assert type(enqueue_val) is int and enqueue_val > 0
  assert type(val_dequeued) is int and val_dequeued > 0
  p =  '<get-register>'
  p +=   eo(reg_to_get)
  p +=   _populate_register_file(
           reg_file     = reg_file,
           reg_file_tmp = reg_file_tmp,
           up_to        = reg_to_get,
           up_to_tmp    = reg_to_get_tmp,
           copy_tmp     = copy_tmp
         )
  p +=   dq(reg_to_get)
  p +=   copy(reg_to_get, reg_to_get_tmp, copy_tmp)
  p +=   '<for-each-reg>'
  p +=     db(reg_to_get_tmp)
  p +=     jf('<get-cur-reg>')
  p +=     jf('<bypass-cur-reg>')
  p +=     jf('<error>')
  p +=     '<bypass-cur-reg>'
  p +=       db(reg_file)
  p +=       jf('<error>')
  p +=       jf('<one>')
  p +=       jf('<sharp>')
  p +=       '<one>'
  p +=         db(reg_file)
  p +=         jf('<error>')
  p +=         jf('<one-one>')
  p +=         jf('<one-sharp>')
  p +=         '<one-one>'
  p +=           eo(reg_file_tmp)
  p +=           eo(reg_file_tmp)
  p +=           jb('<bypass-cur-reg>')
  p +=         '<one-sharp>'
  p +=           eo(reg_file_tmp)
  p +=           es(reg_file_tmp)
  p +=           jb('<bypass-cur-reg>')
  p +=       '<sharp>'
  p +=         db(reg_file)
  p +=         jf('<error>')
  p +=         jf('<error>')
  p +=         jf('<sharp-sharp>')
  p +=         '<sharp-sharp>'
  p +=           es(reg_file_tmp)
  p +=           es(reg_file_tmp)
  p +=           jb('<for-each-reg>')
  p +=     '<get-cur-reg>'
  if False == is_enqueue: # Include instructions for dequeue
    p +=     db(reg_file)
    p +=     jf('<error>')
    p +=     jf('<-one->')
    p +=     jf('<-sharp->')
    p +=     '<-one->'
    p +=       db(reg_file)
    p +=       jf('<error>')
    p +=       jf('<-one-one->')
    p +=       jf('<-one-sharp->')
    p +=       '<-one-one->'
    p +=         eo(val_dequeued)
    p +=         jf('<epilogue>')
    p +=       '<-one-sharp->'
    p +=         es(val_dequeued)
    p +=         jf('<epilogue>')
    p +=     '<-sharp->'
    p +=       db(reg_file)
    p +=       jf('<error>')
    p +=       jf('<error>')
    p +=       jf('<-sharp-sharp->')
    p +=       '<-sharp-sharp->'
    p +=         es(reg_file_tmp)
    p +=         es(reg_file_tmp)
    p +=         jf('<epilogue>')
  elif True == is_enqueue: # Include instructions for enqueue
    p +=     db(reg_file)
    p +=     jf('<error>')
    p +=     jf('<--one-->')
    p +=     jf('<--sharp-->')
    p +=     '<--one-->'
    p +=       db(reg_file)
    p +=       jf('<error>')
    p +=       jf('<--one--one-->')
    p +=       jf('<--one--sharp-->')
    p +=       '<--one--one-->'
    p +=         eo(reg_file_tmp)
    p +=         eo(reg_file_tmp)
    p +=         jb('<get-cur-reg>')
    p +=       '<--one--sharp-->'
    p +=         eo(reg_file_tmp)
    p +=         es(reg_file_tmp)
    p +=         jb('<get-cur-reg>')
    p +=     '<--sharp-->'
    p +=       db(reg_file)
    p +=       jf('<error>')
    p +=       jf('<error>')
    p +=       jf('<--sharp--sharp-->')
    p +=       '<--sharp--sharp-->'
    p +=         copy(enqueue_val, reg_file_tmp, copy_tmp)
    p +=         es(reg_file_tmp)
    p +=         es(reg_file_tmp)
    p +=         jf('<epilogue>')
  else:
    assert False # ``is_enqueue'' must be either True or False
  p +=   '<epilogue>'
  p +=     move(reg_file, reg_file_tmp)
  p +=     move(reg_file_tmp, reg_file)
  p +=   '<error>'
  return labels_to_offsets(p)

def _execute_instruction(
  instr_idx,
  instr_ones,
  instr_ones_tmp,
  instr_sharps,
  reg_file,
  reg_file_tmp,
  copy_tmp,
  enqueue_val_reg,
  val_dequeued_reg
):
  p =  '<execute-instruction>'
  p +=   dq(instr_sharps)
  p +=   db(instr_sharps)
  p +=   jf('<enqueue-one>')
  p +=   jf('<error>')
  p +=     db(instr_sharps)
  p +=     jf('<enqueue-sharp>')
  p +=     jf('<error>')
  p +=       db(instr_sharps)
  p +=       jf('<jump-forward>')
  p +=       jf('<error>')
  p +=         db(instr_sharps)
  p +=         jf('<jump-backward>')
  p +=         jf('<error>')
  p +=           db(instr_sharps)
  p +=           jf('<dequeue-and-branch>')
  p +=           jf('<error>')
  p +=           jf('<error>')
  p +=   '<enqueue-one>'
  p +=     dq(instr_ones)
  p +=     eo(enqueue_val_reg)
  p +=     eo(enqueue_val_reg)
  p +=     _enqueue_or_dequeue(
             reg_file       = reg_file,
             reg_file_tmp   = reg_file_tmp,
             reg_to_get     = instr_ones,
             reg_to_get_tmp = instr_ones_tmp,
             copy_tmp       = copy_tmp,
             is_enqueue     = True,
             enqueue_val    = enqueue_val_reg,
             val_dequeued   = val_dequeued_reg
           )
  p +=     eo(instr_ones)
  p +=     eo(instr_idx) # Advance the program counter by 1
  p +=     jf('<end>')
  p +=   '<enqueue-sharp>'
  p +=     dq(instr_ones)
  p +=     eo(enqueue_val_reg)
  p +=     es(enqueue_val_reg)
  p +=     _enqueue_or_dequeue(
             reg_file       = reg_file,
             reg_file_tmp   = reg_file_tmp,
             reg_to_get     = instr_ones,
             reg_to_get_tmp = instr_ones_tmp,
             copy_tmp       = copy_tmp,
             is_enqueue     = True,
             enqueue_val    = enqueue_val_reg,
             val_dequeued   = val_dequeued_reg
           )
  p +=     eo(instr_ones)
  p +=     eo(instr_idx) # Advance the program counter by 1
  p +=     jf('<end>')
  p +=   '<jump-forward>'
  p +=     copy(instr_ones, instr_idx, copy_tmp)
  p +=     jf('<end>')
  p +=   '<jump-backward>'
  p +=     copy(instr_ones, instr_ones_tmp, copy_tmp)
  p +=     '<jb-loop>'
  p +=       db(instr_ones_tmp)
  p +=       jf('<end>')
  p +=       jf('<jb-one>')
  p +=       jf('<error>')
  p +=       '<jb-one>'
  p +=         db(instr_idx)
  p +=         jf('<error>')
  p +=         jb('<jb-loop>')
  p +=         jf('<error>')
  p +=   '<dequeue-and-branch>'
  p +=     dq(instr_ones)
  p +=     _enqueue_or_dequeue(
             reg_file       = reg_file,
             reg_file_tmp   = reg_file_tmp,
             reg_to_get     = instr_ones,
             reg_to_get_tmp = instr_ones_tmp,
             copy_tmp       = copy_tmp,
             is_enqueue     = False,
             enqueue_val    = enqueue_val_reg,
             val_dequeued   = val_dequeued_reg
           )
  p +=     eo(instr_ones)
  p +=     db(val_dequeued_reg)
  p +=     jf('<db-empty>')
  p +=     jf('<db-one>')
  p +=     jf('<db-sharp>')
  p +=     '<db-empty>'
  p +=       eo(instr_idx)
  p +=       jf('<end>')
  p +=     '<db-one>'
  p +=       eo(instr_idx)
  p +=       eo(instr_idx)
  p +=       jf('<end>')
  p +=     '<db-sharp>'
  p +=       eo(instr_idx)
  p +=       eo(instr_idx)
  p +=       eo(instr_idx)
  p +=       jf('<end>')
  p +=   '<end><error>'
  return labels_to_offsets(p)

def universal(
  prog_txt         = PROG_TXT,
  prog_txt_tmp     = PROG_TXT_TMP,
  instr_idx        = INSTR_IDX,
  instr_idx_tmp    = INSTR_IDX_TMP,
  instr_ones       = INSTR_ONES,
  instr_ones_tmp   = INSTR_ONES_TMP,
  instr_sharps     = INSTR_SHARPS,
  instr_one_save   = INSTR_ONE_SAVE,
  reg_file         = REG_FILE,
  reg_file_tmp     = REG_FILE_TMP,
  enqueue_val_reg  = ENQUEUE_VAL_REG,
  val_dequeued_reg = VAL_DEQUEUED_REG,
  copy_tmp         = COPY_TMP,
  exception_reg    = EXCEPTION_REG
):
  p =  '<program-loop>'
  p +=   _fetch_instruction(
           prog_txt       = prog_txt,
           prog_txt_tmp   = prog_txt_tmp,
           instr_idx      = instr_idx,
           instr_idx_tmp  = instr_idx_tmp,
           instr_ones     = instr_ones,
           instr_sharps   = instr_sharps,
           instr_one_save = instr_one_save,
           copy_tmp       = copy_tmp
         )
         # Check for terminal state, otherwise execute instruction
  p +=   db(instr_sharps)
  p +=   jf('<epilogue>') # jf('<end>') for debugging
  p +=   jf('<error>')
  p +=     es(instr_sharps)
  p +=     _execute_instruction(
             instr_idx        = instr_idx,
             instr_ones       = instr_ones,
             instr_ones_tmp   = instr_ones_tmp,
             instr_sharps     = instr_sharps,
             reg_file         = reg_file,
             reg_file_tmp     = reg_file_tmp,
             copy_tmp         = copy_tmp,
             enqueue_val_reg  = enqueue_val_reg,
             val_dequeued_reg = val_dequeued_reg
           )
  p +=     clear(instr_ones)
  p +=     clear(enqueue_val_reg)
  p +=     clear(val_dequeued_reg)
  p +=     jb('<program-loop>') # jf('<end>') for debugging
  p +=   '<epilogue>'
  p +=     clear(instr_idx)
           # Write the encoded value of R1 into the actual program txt reg (R1)
  p +=     clear(instr_ones)
  p +=     clear(prog_txt)
  for _ in range(prog_txt-1):
    p +=   eo(instr_ones)
  p +=     '<final-output-loop>'
  p +=       _enqueue_or_dequeue(
               reg_file       = reg_file,
               reg_file_tmp   = reg_file_tmp,
               reg_to_get     = instr_ones,
               reg_to_get_tmp = instr_ones_tmp,
               copy_tmp       = copy_tmp,
               is_enqueue     = False,
               enqueue_val    = enqueue_val_reg,
               val_dequeued   = val_dequeued_reg
             )
  p +=       db(val_dequeued_reg)
  p +=       jf('<check-reg-remaining>')
  p +=       jf('<one>')
  p +=       jf('<sharp>')
  p +=       '<one>'
  p +=         eo(prog_txt)
  p +=         jb('<final-output-loop>')
  p +=       '<sharp>'
  p +=         es(prog_txt)
  p +=         jb('<final-output-loop>')
  p +=       '<check-reg-remaining>'
               # If the reg file encodes only empty registers, we want to clear
               #   it out for a good halt, but if there are any non-empty regs,
               #   we want to leave them so we get a bad halt
  p +=         db(reg_file)
  p +=         jf('<end>')
  p +=         jf('<-one->')
  p +=         jf('<-sharp->')
  p +=         '<-one->'
  p +=           eo(exception_reg)
  p +=           jf('<end>')
  p +=         '<-sharp->'
  p +=           db(reg_file)
  p +=           jf('<error>')
  p +=           jf('<error>')
  p +=           jb('<check-reg-remaining>')
  p +=   '<end><error>'
  return labels_to_offsets(p)



from ..tools.instruction_set import enqueue_one as eo
from ..tools.instruction_set import enqueue_sharp as es
from ..tools.instruction_set import jump_forward as jf
from ..tools.instruction_set import jump_backward as jb
from ..tools.instruction_set import dequeue_and_branch as db
from ..tools.labels_to_offsets import labels_to_offsets
from ..programs.clear import clear
from ..programs.move import move
from ..programs.copy import copy
from ..programs.add_unary import add_unary

# mult_unary
# ==========
# regA    : positive integer, the index of the register containing the
#         :   first factor and the register in which the product will be written
#         :   at the end of the computation
# regB    : positive integer, the index of the register containing the second
#         :   factor which will be returned empty
# tmpA    : positive integer, the first temporary register which should be
#         :   passed empty and will be returned empty
# tmpB    : positive integer, the second temporary register which should be
#         :   passed empty and will be returned empty
# tmpCopy : positive integer, the third temporary register which should be
#         :   passed empty and will be returned empty
# returns : string, the 1# program with that when executed will multiply the
#         :   unary numbers contained in regA and regB, writing the result into
#         :   regA and leaving regB empty
def mult_unary(regA=1, regB=2, tmpA=3, tmpB=4, tmpCopy=5):
  p =  '<mult_loop>'
  p +=   db(regA)
  p +=   jf('<empty>')
  p +=   jf('<one>')
  p +=   jf('<sharp>')
  p +=   '<empty>'
  p +=     jf('<epilogue>')
  p +=   '<one>'
  p +=     copy(regB, tmpB, tmpCopy)
  p +=     add_unary(tmpA, tmpB)
  p +=     jb('<mult_loop>')
  p +=   '<sharp>'
  p +=     jf('<error>')
  p += '<epilogue>'
  p +=   move(tmpA, regA)
  p +=   clear(regB)
  p += '<error>'
  return labels_to_offsets(p)



from ..tools.instruction_set import enqueue_one as eo
from ..tools.instruction_set import enqueue_sharp as es
from ..tools.instruction_set import jump_forward as jf
from ..tools.instruction_set import jump_backward as jb
from ..tools.instruction_set import dequeue_and_branch as db
from ..tools.labels_to_offsets import labels_to_offsets
from ..programs.move import move

# diag
# ====
# reg     : positive integer, the index of the register on whose contents the
#         :   the diag program will be executed and to which the output
#         :   write-program-text will be prepended
# tmpA    : positive integer, the index of the first temporary register
# tmpB    : positive integer, the index of the second temporary register
# returns : string, the 1# program with the property that when it is run with
#         :   contents ``x'' in register ``reg'' and registers ``tmpA'' and
#         :   ``tmpB'' empty, a halt is eventually reached with contents ``y''
#         :   in regsiter ``reg'' and registers ``tmpA'' and ``tmpB'' empty,
#         :   where ``y'' is equal to phi_{write}(x) + x
def diag(reg, tmpA, tmpB):
  p =  '<top>'
  p +=   db(reg)
  p +=   jf('<empty>')
  p +=   jf('<one>')
  p +=   '<sharp>'
  for _ in range(reg):
    p +=   eo(tmpA)
  p +=     es(tmpA)
  p +=     es(tmpA)
  p +=     es(tmpB)
  p +=     jb('<top>')
  p +=   '<one>'
  for _ in range(reg):
    p +=   eo(tmpA)
  p +=     es(tmpA)
  p +=     eo(tmpB)
  p +=     jb('<top>')
  p +=   '<empty>'
  p +=     move(tmpA, reg)
  p +=     move(tmpB, reg)
  return labels_to_offsets(p)



from ..interpreter.interpreter import parse, unparse

def findme(pair,pair_list): 
  x = [i for i in range(len(pair_list)) if pair_list[i][0]==pair[1]]
  return x[0]

def add_placeholder(line):
  keywords = ['cases','add1','add#','goto','non_label']
  if line[0] in keywords:
    return(['non_label']+line)
  elif line[0][0] in ['1','#']:
    return(['non_label']+line)  
  else:
    return(line)

def flatten(tuple):
  b = len(tuple)
  if b > 1 and tuple[1] == "cases":
    n = tuple[2]
    x = ones(n)+'#####'
    return [[tuple[0],x], [tuple[0]+'@1',tuple[3]], [tuple[0]+'@2',tuple[4]], [tuple[0]+'@3',tuple[5]]]
  elif b > 1 and tuple[1]=='add1':
    return([[tuple[0],ones(tuple[2])+'#']])
  elif b > 1 and tuple[1]=='add#':
    return([[tuple[0],ones(tuple[2])+'##']])
  elif b > 1 and tuple[1]=='goto':
    return [[tuple[0],tuple[2]]]
  elif tuple[1][0] in ['1', '#']:
    k = parse(tuple[1])
    m = len(k)
    return([[tuple[0],k[j]] for j in range(m)])


def ones(n):
  return(unparse(['1' for i in range(n)])) 

def resolve(index,pair_list):
  pair_in_list = pair_list[index]
  #print("pair_in_list " + str(pair_in_list))
  first_char = pair_in_list[1][0]
  #print(first_char)
  if first_char == '1' or first_char == '#':
    return pair_in_list[1]
  elif pair_in_list[1]=='end':
    n = len(pair_list) - index
    return(ones(n)+'###')
  else:
    k = findme(pair_in_list,pair_list)
    if k > index:
      return(ones(k-index)+'###')
    if k < index:
      return(ones(index-k)+'####')


def sanity(line_list):
  w = [add_placeholder(line) for line in line_list]
  #print("[add_placeholder(line) for line in line_list] is")
  #print(w)
  s1 = [flatten(line) for line in w]
  #print('[flatten(line) for line in w] is')
  #print(s1)
  t1 = [item for sublist in s1 for item in sublist]
  #print('[item for sublist in s1 for item in sublist] is')
  #print(t1)
  n = len(t1)
  #print('len of t1 = ' + str(n))
  u1 = [resolve(i,t1) for i in range(n)] 
  return(unparse(u1))



import sys

# labels_to_offsets
# =================
# program : string, the ``label-enhanced'' 1# program
# lstart  : character, the label-start delimiter, default is '<'
# lend    : character, the label-end delimiter, default is '>'
# returns : string, the program with its marker labels removed and its
#         :   instruction-argument labels translated to their relevant offsets
#
#   example
#   -------
#     >>> clear_1_using_labels = \
#     ...   '<start>'+\
#     ...     '1#####'+\
#     ...     '<end>###'+\
#     ...     '<start>####'+\
#     ...     '<start>####'+\
#     ...   '<end>'
#     >>> clear_1 = labels_to_offsets( clear_1_using_labels )
#     >>> print(clear_1)
#     '1#####111###11####111####'
#
#
def labels_to_offsets(program, lstart='<', lend='>'):
  program = ''.join( program.split() ) # Remove all whitespace
  assert lstart not in ['1','#']
  assert lend not in ['1','#']
  assert lstart != lend # The label start and end delimiters must differ
  # Phase 1
  # Parse the label-containing program
  parse_of_program = []
  prog_len_in_chars = len(program)
  label__instruction_index = dict()
  char_index = 0
  if char_index < prog_len_in_chars:
    cur_char = program[char_index]
  while char_index < prog_len_in_chars:
    # Phase 1 Step 1
    # Parse label, if present
    assert cur_char in [lstart,'1']
    if cur_char == lstart:
      label = ''
      while True:
        char_index += 1
        assert char_index < prog_len_in_chars
        cur_char = program[char_index]
        assert cur_char not in ['1','#',lstart] # A label cannot contain these
        if cur_char == lend:
          break
        label += cur_char
      char_index += 1
      if char_index < prog_len_in_chars:
        cur_char = program[char_index]
      else:
        cur_char = 'EOF'
      if cur_char == 'EOF' or cur_char == lstart or cur_char == '1':
        assert label not in label__instruction_index
        label__instruction_index[label] = len(parse_of_program)
        if cur_char == 'EOF':
          break # Phase 1 complete
        elif cur_char == lstart:
          continue # Parse the next label
      elif cur_char == '#':
        parse_of_program.append([label])
      else:
        assert False # Unexpected character followed the label-end delimiter
    # Phase 1 Step 2
    # Parse the instruction argument, if present
    arg = ''
    while cur_char == '1':
      arg += '1'
      char_index += 1
      assert char_index < prog_len_in_chars # Unexpected end of program
      cur_char = program[char_index]
    if len(arg) > 0:
      parse_of_program.append([arg])
    # Phase 1 Step 3
    # Parse the opcode
    opcode = ''
    assert cur_char == '#' # Unexpected character followed instruction argument
    while cur_char == '#':
      opcode += '#'
      if len(opcode) == 6:
        assert False # Opcode can be five sharps at most
      char_index += 1
      if char_index == prog_len_in_chars:
        break
      cur_char = program[char_index]
    assert len(parse_of_program[-1]) == 1
    parse_of_program[-1].append(opcode)
  # Phase 2
  # Replace the labels with offsets
  program = ''
  for instruction_index, arg__opcode in enumerate(parse_of_program):
    arg = arg__opcode[0]
    opcode = arg__opcode[1]
    if arg in label__instruction_index:
      label_ii = label__instruction_index[arg]
      offset = label_ii - instruction_index
      if offset < 0:
        if opcode != '####':
          sys.stderr.write('Offending label: '+arg)
          assert False # We should be jumping backward, but we aren't
        offset = -offset
      elif offset > 0:
        assert opcode == '###'
      else:
        assert offset == 0 # If fails, bug internal to function
        assert False # Cannot jump to self
      arg = ''.join(['1' for _ in range(offset)])
    # Confirm we have a valid arg; that is, we don't have an undefined label
    for char in arg:
      if char != '1':
        sys.stderr.write('offsets_to_labels: encountered undefined label: '+arg)
        assert False
    program += arg + opcode
  return program



from ..tools.ones import ones
from ..programs.move import move

# copy
# ====
# src     : positive integer, the index of the source register
# dst     : positive integer, the index of the destination register
# tmp     : positive integer, the index of the temporary register
# returns : string, the 1# program to copy the contents from register ``src'' to
#         :  register ``dst'' using register ``tmp'' as temporary scratch space
def copy(src, dst, tmp):
  p = ones(src)+'#####' # Cases on register ``src''
  p += '11111 111###'   # Go forward eight 
  p += '1111###'        # Go forward four
  p += ones(dst)+'##'   # Append ``#'' to register ``dst''
  p += ones(tmp)+'##'   # Append ``#'' to register ``tmp''
  p += '11111####'      # Go backward five
  p += ones(dst)+'#'    # Append ``1'' to register ``dst''
  p += ones(tmp)+'#'    # Append ``1'' to register ``tmp''
  p += '11111 111####'  # Go backward eight
  p += move(tmp, src)   # Move the contents of register ``tmp'' to register
                        #   ``src''
  return p



# ones
# ====
# n       : positive integer, the length of the ones string to return
# returns : string, a string containing exactly ``n'' ones
#
#   example
#   -------
#     >>> print(ones(3))
#     '111'
#
def ones(n):
  assert type(n) == int # ``n'' must be an integer
  assert n > 0 # ``n'' must be positive
  return ''.join(['1' for _ in range(n)])



from ..tools.ones import ones

def is_positive_integer(n):
  return type(n) == int and n > 0

def is_string(s):
  return type(s) == str

def enqueue_one(register):
  assert is_positive_integer(register)
  return ones(register) + '#'

def enqueue_sharp(register):
  assert is_positive_integer(register)
  return ones(register) + '##'

def jump_forward(offset):
  if is_positive_integer(offset):
    return ones(offset) + '###'
  elif is_string(offset):
     return offset + '###'
  else:
     assert False # ``offset'' must be a positive integer or label string

def jump_backward(offset):
  if is_positive_integer(offset):
    return ones(offset) + '####'
  elif is_string(offset):
    return offset + '####'
  else:
    assert False # ``offset'' must be a positive integer or label string

def dequeue_and_branch(register):
  assert is_positive_integer(register)
  return ones(register) + '#####'

def no_operation():
  return jump_forward(1)

def dequeue(register):
  p =  dequeue_and_branch(register)
  p += no_operation()
  p += no_operation()
  return p



from ..tools.instruction_set import enqueue_one as eo
from ..tools.instruction_set import enqueue_sharp as es
from ..tools.instruction_set import jump_forward as jf
from ..tools.instruction_set import jump_backward as jb
from ..tools.instruction_set import dequeue_and_branch as db
from ..tools.instruction_set import dequeue as dq
from ..tools.labels_to_offsets import labels_to_offsets
from ..tools.sanity import sanity
from ..programs.clear import clear
from ..programs.move import move
from ..programs.copy import copy

# _beige
# ======
#   Design and implementation of _beige function by Sarah Peterson
#   The function is used to fetch the instruction in the program text (R1)
#     that is pointed to by the program counter (R2), with the fetched
#     instruction returned in R3, and R1 and R2 remain as they were before
#     the call
def _beige():
  p = '11#####111111111111111111111111111111111111111111111111111111111###11###1#111111111#111#####111###11####111####1#####1111111111111111111111111111111###11###1#111#1#####11111111111111111111111111###11###111###111#11111####111##1#####1111111111111111111###111###111##1111####111111#1#####111111###111###111111##1111####111111#111111####111111#####111111###111###1##1111####1#111111####111#####11111111###1111###11111##11111111##11111####11111#11111111#11111111####11111111#####111111###111###111##1111####111#111111####111111111111111111111111111111111111111111111111111111111####1#####111111###111###11111##1111####11111#111111####11111#####111111###111###1##1111####1#111111####111111111#####111111###111###11##1111####11#111111####'
  return p

# _jump
# =====
#   Design and implementation of _jump function by Joshua Kim
#   The function is used to execute instructions with opcode of three and four
#     sharps (i.e., jump forward or backward)
def _jump():
  p = [
    # This program assumes that it get the nth instruction in R3 (e.g. 1111### in R3)
    # R1 has the whole program text.
    # R2 has the currently instruction number. (what instruction we are on).
    # R3 has the nth instruction.
    # R4 is emtpy.
    
    #one_loop will count the number of 1's in the instruction and put them in R4.
    ['one_loop', 'cases', 3, 'empty_one', 'one_one', 'sharp_one'],
      ['empty_one', 'goto', 'end'],  # If empty go to end
      ['one_one', '1111#'], # If one, add 1 in R4 and one_loop again.
      ['goto', 'one_loop'], 
      ['sharp_one', 'goto', 'sharp_loop_one'], # After counting all the 1's, how count the sharps.
    
    
    # This loop will count how many sharps are in the instruction.
    ['sharp_loop_one', 'cases', 3, 'empty__one', 'one__one', 'sharp__one'],
      ['empty__one', 'goto', 'one_sharp_case'], # If it is empty, instruction had one sharp, so go to one sharp cases which is adding a 1 in a register.
      ['one__one', 'goto', 'end'], # There can't be a one, so if it is a one, end.
      ['sharp__one', 'goto', 'sharp_loop_two'], # If there is a sharp, then we would go to next loop to see how many more sharps there are.
    
    # Same logic, this loop will count sharps. This one is for 2 sharps.
    ['sharp_loop_two', 'cases', 3, 'empty__two', 'one__two', 'sharp__two'],
      ['empty__two', 'goto', 'two_sharp_case'],
      ['one__two', 'goto', 'end'],
      ['sharp__two', 'goto', 'sharp_loop_three'],
    
    # This one is three sharps which is important.
    ['sharp_loop_three', 'cases', 3, 'empty__three', 'one__three', 'sharp__three'],
      ['empty__three', 'goto', 'three_sharp_case'], 
      ['one__three', 'goto', 'end'],
      ['sharp__three', 'goto', 'sharp_loop_four'], 
    
    # This one is four sharp.
    ['sharp_loop_four', 'cases', 3, 'empty__four', 'one__four', 'sharp__four'],
      ['empty__four', 'goto', 'four_sharp_case'],
      ['one__four', 'goto', 'end'],
      ['sharp__four', 'goto', 'sharp_loop_five'],
    
    # This one is five sharp.
    ['sharp_loop_five', 'cases', 3, 'empty__five', 'one__five', 'sharp__five'],
      ['empty__five', 'goto', 'five_sharp_case'],
      ['one__five', 'goto', 'end'],
      ['sharp__five', 'goto', 'end'],
    
    # This is what happens for each sharps.
    ['one_sharp_case', 'goto', 'end'],  # This is where you would put adding a 1 (one sharp case)
    ['two_sharp_case', 'goto', 'end'],  # Two sharp case
    
    
    ['three_sharp_case', move(4,2)], # Three sharp case, this is jump foward, so add the 1's we counted from the instruction to R2 which has the number of current instruction we are on.
    
    # This is the jump backward case.
    ['four_sharp_case', 'goto', 'jump_back'], # we will run cases on R4, and each time delete a 1 from R2 untill R4 is empty.
    ['jump_back', 'cases', 4, 'empty_jb', 'one_jb', 'sharp_jb'],
      ['empty_jb', 'goto', 'end'],
      ['one_jb', 'goto', 'minus_one_cases'],
      ['sharp_jb', 'goto', 'end'],
    
    # This is where deleting a 1 from R2 occurs.
    ['minus_one_cases', 'cases', 2, 'empty_MO', 'one_MO', 'sharp_mo'],
      ['empty_MO', 'goto', 'jump_back'],
      ['one_MO', 'goto', 'jump_back'],
      ['sharp_mo', 'goto', 'jump_back'],
    
    ['five_sharp_case', 'goto', 'end'],
    
    ['end', '1###']
  ]
  return sanity(p)

# _populate_register_file
# =======================
#   Design concept of _populate_register_file function by Rhodes Talaga and
#     Jacob Striebel; implementation by Jacob Striebel
def _populate_register_file(
  reg_file,
  reg_file_tmp,
  up_to,
  up_to_tmp,
  copy_tmp
):
  p =  copy(up_to, up_to_tmp, copy_tmp)
  p += '<populate-register-file>'
  p +=   db(up_to_tmp)
  p +=   jf('<epilogue>')
  p +=   jf('<check-reg-init-loop>')
  p +=   jf('<error>')
  p +=   '<check-reg-init-loop>'
  p +=     db(reg_file)
  p +=     jf('<handle-empty>')
  p +=     jf('<handle-one>')
  p +=     jf('<handle-sharp>')
  p +=     '<handle-one>'
  p +=       eo(reg_file_tmp)
  p +=       db(reg_file)
  p +=       jf('<error>')
  p +=       jf('<handle-one-one>')
  p +=       jf('<handle-one-sharp>')
  p +=       '<handle-one-one>'
  p +=         eo(reg_file_tmp)
  p +=         jb('<check-reg-init-loop>')
  p +=       '<handle-one-sharp>'
  p +=         es(reg_file_tmp)
  p +=         jb('<check-reg-init-loop>')
  p +=     '<handle-sharp>'
  p +=       es(reg_file_tmp)
  p +=       db(reg_file)
  p +=       jf('<error>')
  p +=       jf('<error>')
  p +=         es(reg_file_tmp)
  p +=         jb('<populate-register-file>')
  p +=     '<handle-empty>'
  p +=       es(reg_file_tmp)
  p +=       es(reg_file_tmp)
  p +=       jb('<populate-register-file>')
  p +=   '<epilogue>'
  p +=     move(reg_file, reg_file_tmp)
  p +=     move(reg_file_tmp, reg_file)
  p +=   '<error>'
  return labels_to_offsets(p)

# _enqueue_or_dequeue
# ===================
#   Design concept of _enqueue_or_dequeue function by J. Philip Pampreen and
#     Jacob Striebel; this implementation by Jacob Striebel
#                :
#                :
# reg_file       : positive integer, the index of the register in which all the
#                :   registers of the program we are executing via the
#                :   universal program are encoded; we expect the registers
#                :   to be encoded using the scheme defined in the universal
#                :   program chapter of the TRM web text by L. Moss
#                :
# reg_file_tmp   : positive integer, the index of the temporary register used
#                :   when cycling through the register-file register
#                :
# reg_to_get     : positive integer, the index of the register that we are
#                :   either enqueuing a character into or dequeuing a character
#                :   from
#                :
# reg_to_get_tmp : positive integer, the index of the register that we use to
#                :   to save the contents of the ``reg_to_get'' register while
#                :   we are doing processing directly on that register
#                :
# copy_tmp       : positive integer, the index of the register that we use for
#                :   temporary scratch space during copy operations
#                :
# is_enqueue     : boolean, whether to return the 1# code module for enqueue
#                :   or dequeue; if ``True'' then enqueue, and if ``false''
#                :   then dequeue; this is the only variable in this entire file
#                :   (universal.py) which is a sort of ``meta'' variable with
#                :   respect to the 1# program that is being generated and which
#                :   isn't included in the 1# program itself; all other
#                :   variables in this file are positive integers which each
#                :   refer to the index of a Moss machine register that is
#                :   operated on in the actual 1# program that is generated
#                :   by the Python code in this file
#                :
# enqueue_val    : positive integer, the index of the register into which the
#                :   the value is passed to this module to enqueue into the
#                :   register file in the case of an enqueue operation; in the
#                :   case of a dequeue operation, this register will not be used
#                :
# val_dequeued   : positive integer, the index of the register into which the
#                :   the value will be stored that is dequeued from the register
#                :   file in the case of a dequeue operation; in the case of an
#                :   enqueue operation, this register will not be used
#                :
# returns        : string, the 1# subprogram of the universal program which will
#                :   either enqueue a character into a specified register in the
#                :   the register file or will dequeue a character from a
#                :   specified register in the register file, depending on the
#                :   value specified for the ``is_enqueue'' argument
#                :
#   Summary of the 1# code modules that are returned for enqueuing and dequeuing
#   ----------------------------------------------------------------------------
#           The enqueue and dequeue modules both begin with a call to the
#   ``_populate_regsiter_file'' module for which the index of the register that
#   will be enqueued to or dequeued from is passed. The result of this call is
#   that the register file is initialized up to the index of the register that
#   will be enqueued to or dequeued from, if it is not already. For example,
#   if the register file contains ``11 ## ##'', encoding an R1 contents of
#   ``1'', an empty R2, and the remaining registers uninitialized, and if
#   further the enqueue or dequeue operation that we are preparing to perform
#   will operate on R4, then our call to the populate-register-file module will
#   result in R3 and R4 being initialized in the regsiter file so the new
#   contents are ``11 ## ## ## ##''. Initializing the register in the register
#   file that we will operate on before we begin the operation simplifies coding
#   the operation itself in 1#.
#           After the register file has been initialized up to the register we
#   will operate on, we next cycle through the register file until we reach
#   the index of that register to operate on. Once the register has been
#   reached, the action is of course different for an enqueue versus a dequeue
#   operation. For enqueue, we continue to cycle through the current register
#   until we reach the end-of-register code ``##''. At this point we append the
#   value that we are enqueuing onto the cycle-temporary-register
#   ``reg_file_tmp'', then we copy the remaining contents of the register file
#   onto the cycle register, followed by copying the cycle register back into
#   the ``reg_file'' register. At this point the enqueue operation is complete.
#           In the case of a dequeue operation, the value encoded into the
#   register of interest in the first position is read, without copying it into
#   the cycle register. Instead it is copied into the ``val_dequeued'' register
#   to be returned to the calling module. After this operation is complete,
#   the remaining contents, if any, of the register that we just dequeued from
#   are copied onto the cycle register, and after that all of the remaining
#   contents of the register file are copied onto the cycle register, and
#   finally the contents of the cycle register are moved back into the register
#   file, to complete the cycle. At this point the dequeue operation is
#   complete.
#
def _enqueue_or_dequeue(
  reg_file,
  reg_file_tmp,
  reg_to_get,
  reg_to_get_tmp,
  copy_tmp,
  is_enqueue,
  enqueue_val,
  val_dequeued
):
  assert type(reg_file) is int and reg_file > 0
  assert type(reg_file_tmp) is int and reg_file_tmp > 0
  assert type(reg_to_get) is int and reg_to_get > 0
  assert type(reg_to_get_tmp) is int and reg_to_get_tmp > 0
  assert type(copy_tmp) is int and copy_tmp > 0
  assert type(is_enqueue) is bool and is_enqueue in [True,False]
  assert type(enqueue_val) is int and enqueue_val > 0
  assert type(val_dequeued) is int and val_dequeued > 0
  p =  '<get-register>'
  p +=   eo(reg_to_get)
  p +=   _populate_register_file(
           reg_file     = reg_file,
           reg_file_tmp = reg_file_tmp,
           up_to        = reg_to_get,
           up_to_tmp    = reg_to_get_tmp,
           copy_tmp     = copy_tmp
         )
  p +=   dq(reg_to_get)
  p +=   copy(reg_to_get, reg_to_get_tmp, copy_tmp)
  p +=   '<for-each-reg>'
  p +=     db(reg_to_get_tmp)
  p +=     jf('<get-cur-reg>')
  p +=     jf('<bypass-cur-reg>')
  p +=     jf('<error>')
  p +=     '<bypass-cur-reg>'
  p +=       db(reg_file)
  p +=       jf('<error>')
  p +=       jf('<one>')
  p +=       jf('<sharp>')
  p +=       '<one>'
  p +=         db(reg_file)
  p +=         jf('<error>')
  p +=         jf('<one-one>')
  p +=         jf('<one-sharp>')
  p +=         '<one-one>'
  p +=           eo(reg_file_tmp)
  p +=           eo(reg_file_tmp)
  p +=           jb('<bypass-cur-reg>')
  p +=         '<one-sharp>'
  p +=           eo(reg_file_tmp)
  p +=           es(reg_file_tmp)
  p +=           jb('<bypass-cur-reg>')
  p +=       '<sharp>'
  p +=         db(reg_file)
  p +=         jf('<error>')
  p +=         jf('<error>')
  p +=         jf('<sharp-sharp>')
  p +=         '<sharp-sharp>'
  p +=           es(reg_file_tmp)
  p +=           es(reg_file_tmp)
  p +=           jb('<for-each-reg>')
  p +=     '<get-cur-reg>'
  if False == is_enqueue: # Include instructions for dequeue
    p +=     db(reg_file)
    p +=     jf('<error>')
    p +=     jf('<-one->')
    p +=     jf('<-sharp->')
    p +=     '<-one->'
    p +=       db(reg_file)
    p +=       jf('<error>')
    p +=       jf('<-one-one->')
    p +=       jf('<-one-sharp->')
    p +=       '<-one-one->'
    p +=         eo(val_dequeued)
    p +=         jf('<epilogue>')
    p +=       '<-one-sharp->'
    p +=         es(val_dequeued)
    p +=         jf('<epilogue>')
    p +=     '<-sharp->'
    p +=       db(reg_file)
    p +=       jf('<error>')
    p +=       jf('<error>')
    p +=       jf('<-sharp-sharp->')
    p +=       '<-sharp-sharp->'
    p +=         es(reg_file_tmp)
    p +=         es(reg_file_tmp)
    p +=         jf('<epilogue>')
  elif True == is_enqueue: # Include instructions for enqueue
    p +=     db(reg_file)
    p +=     jf('<error>')
    p +=     jf('<--one-->')
    p +=     jf('<--sharp-->')
    p +=     '<--one-->'
    p +=       db(reg_file)
    p +=       jf('<error>')
    p +=       jf('<--one--one-->')
    p +=       jf('<--one--sharp-->')
    p +=       '<--one--one-->'
    p +=         eo(reg_file_tmp)
    p +=         eo(reg_file_tmp)
    p +=         jb('<get-cur-reg>')
    p +=       '<--one--sharp-->'
    p +=         eo(reg_file_tmp)
    p +=         es(reg_file_tmp)
    p +=         jb('<get-cur-reg>')
    p +=     '<--sharp-->'
    p +=       db(reg_file)
    p +=       jf('<error>')
    p +=       jf('<error>')
    p +=       jf('<--sharp--sharp-->')
    p +=       '<--sharp--sharp-->'
    p +=         copy(enqueue_val, reg_file_tmp, copy_tmp)
    p +=         es(reg_file_tmp)
    p +=         es(reg_file_tmp)
    p +=         jf('<epilogue>')
  else:
    assert False # ``is_enqueue'' must be either True or False
  p +=   '<epilogue>'
  p +=     move(reg_file, reg_file_tmp)
  p +=     move(reg_file_tmp, reg_file)
  p +=   '<error>'
  return labels_to_offsets(p)

# universal_alt
# =============
#   Design concept credit for the universal_alt function goes to all group
#     members; the implementation below is by Jacob Striebel; credit for each
#     helper function that is defined above in this file and that is
#     used by the universal_alt function is given in the helper
#     functions' header comments
def universal_alt(
  prog_txt         = 1,  # Program text : the index of the register that contains
                         #   the text of the 1# program that we are running the
                         #   universal program on
  instr_idx        = 2,  # Instruction index : the index of the register that
                         #   contains the index of the instruction in the
                         #   the program text which we would like to execute or
                         #   which is in the process of being executed; this value,
                         #   in addition to being call the ``instruction index'',
                         #   can synonymously be called the ``program counter'' and
                         #   is encoded in unary, with an empty register meaning
                         #   zero (i.e., the first instruction in the program),
                         #   ``1'' meaning one, ``11'' meaning two, etc.)
  instr_txt        = 3,  # Instruction text : the index of the register which is
                         #   used to store the text of an instruction that has
                         #   just been fetched, before the instruction is split
                         #   into its ones and sharps parts
  jump_tmp         = 4,  # Jump temporary : the index of the jump temporary
                         #   register which is used in the _jump function
  instr_ones       = 20, # Instruction ones : the index of the instruction ones
                         #   register; the process of fetching an instruction
                         #   writes the number of ones in that instruction into
                         #   the instruction ones register; e.g., after the
                         #   instruction ``1111###'' is fetched, the contents of
                         #   this register will be ``1111''
  instr_ones_tmp   = 21, # Instruction ones temporary : the index of the register
                         #   used to save the instruction ones register when we
                         #   are doing operations directly on that register
  instr_sharps     = 22, # Instruction sharps : the index of the instruction sharps
                         #   register; the process of fetching an instruction
                         #   writes the number of sharps in that instruction into
                         #   the instruction sharps register; e.g., after the
                         #   instruction ``1111###'' is fetched, the contents of
                         #   this register will be ``###''
  reg_file         = 23, # Register file : the index of the register into which all
                         #   the registers of the program we are executing are
                         #   encoded via the encoding defined in the TRM web
                         #   text by L. Moss
  reg_file_tmp     = 24, # Register file temporary : the index of the temporary
                         #   register used when cycling through the contents of the
                         #   register file
  enqueue_val_reg  = 25, # Enqueue value regsiter : the index of the register into
                         #   which the value to enqueue is written before control
                         #   is transfered to the enqueue module
  val_dequeued_reg = 26, # Value dequeued register : the index of the register into
                         #   which the value dequeued from the register file is
                         #   written before a call to the dequeue module returns
  copy_tmp         = 27  # Copy temporary : the index of the temporary register
                         #   that used as scratch space during copy operations
):
  p =      eo(instr_idx) # Initialize the program counter PC (i.e., R2) to point
                         #   to the first instruction
  p += '<program_loop>'
  p +=     _beige()    # Fetch the instruction pointed to by PC and write it
                       #   into R3
  # Execute the instruction in R3; if R3 is empty, we are in the terminal state
  p +=     db(instr_txt)
  p +=     jf('<instruction_parse_precheck_empty>')
  p +=     jf('<instruction_parse_precheck_one>')
  p +=     jf('<instruction_parse_precheck_sharp>')
  p +=     '<instruction_parse_precheck_empty>'
               # The instruction register (R3) is empty
  p +=         jf('<epilogue>')
  p +=     '<instruction_parse_precheck_one>'
               # The instruction register (R3) starts with a one: so far so good
  p +=         eo(instr_ones)
  p +=         jf('<instruction_get_ones_loop>')
  p +=     '<instruction_parse_precheck_sharp>'
               # Error: an instruction cannot begin with a sharp
  p +=         jf('<end>')
  p +=     '<instruction_get_ones_loop>'
  p +=         db(instr_txt)
  p +=         jf('<instruction_get_ones_empty>')
  p +=         jf('<instruction_get_ones_one>')
  p +=         jf('<instruction_get_ones_sharp>')
  p +=         '<instruction_get_ones_empty>'
                   # Error: the data part of an instruction must be followed by
                   #   must be followed by a sharp
  p +=             jf('<end>')
  p +=         '<instruction_get_ones_one>'
  p +=             eo(instr_ones)
  p +=             jb('<instruction_get_ones_loop>')
  p +=         '<instruction_get_ones_sharp>'
  p +=             es(instr_sharps)
  p +=             jf('<instruction_get_sharps_loop>')
  p +=     '<instruction_get_sharps_loop>'
  p +=         db(instr_txt)
  p +=         jf('<instruction_get_sharps_empty>')
  p +=         jf('<instruction_get_sharps_one>')
  p +=         jf('<instruction_get_sharps_sharp>')
  p +=         '<instruction_get_sharps_empty>'
  p +=             jf('<execute_instruction>')
  p +=         '<instruction_get_sharps_one>'
                   # Error: in a single instruction, a one cannot follow a
                   #   sharp
  p +=             jf('<end>')
  p +=         '<instruction_get_sharps_sharp>'
  p +=             es(instr_sharps)
  p +=             jb('<instruction_get_sharps_loop>')
  p +=     '<execute_instruction>'
           # We first save the full instruction text back into its regsiter
           #   instr_txt (R3) which we just clobbered when we separated data
           #   (ones) from opcode (sharps)
  p +=     copy(instr_ones, instr_txt, copy_tmp)
  p +=     copy(instr_sharps, instr_txt, copy_tmp)
           # Decode and execute
  p +=     db(instr_sharps)
  p +=     jf('<instruction_read_first_sharp_empty>')
  p +=     jf('<instruction_read_first_sharp_one>')
  p +=     jf('<instruction_read_first_sharp_sharp>')
  p +=     '<instruction_read_first_sharp_empty>'
               # Error: a legal instruction opcode cannot contain zero sharps
  p +=         jf('<end>')
  p +=     '<instruction_read_first_sharp_one>'
               # Error
  p +=         jf('<end>')
  p +=     '<instruction_read_first_sharp_sharp>'
  p +=         jf('<instruction_read_second_sharp>')
  p +=     '<instruction_read_second_sharp>'
  p +=     db(instr_sharps)
  p +=     jf('<instruction_read_second_sharp_empty>')
  p +=     jf('<instruction_read_second_sharp_one>')
  p +=     jf('<instruction_read_second_sharp_sharp>')
  p +=     '<instruction_read_second_sharp_empty>'
               ############################################
               # Opcode is a single sharp (enqueue a one) #
               ############################################
  p +=         dq(instr_ones)
  p +=         eo(enqueue_val_reg)
  p +=         eo(enqueue_val_reg)
  p +=         _enqueue_or_dequeue(
                 reg_file       = reg_file,
                 reg_file_tmp   = reg_file_tmp,
                 reg_to_get     = instr_ones,
                 reg_to_get_tmp = instr_ones_tmp,
                 copy_tmp       = copy_tmp,
                 is_enqueue     = True,
                 enqueue_val    = enqueue_val_reg,
                 val_dequeued   = val_dequeued_reg
               )
  p +=         clear(instr_ones) #eo(instr_ones)
  p +=         clear(enqueue_val_reg)
  p +=         eo(instr_idx) # Advance the program counter by one
  p +=         jb('<program_loop>')
               ############################################
               ############################################
  p +=     '<instruction_read_second_sharp_one>'
               # Error
  p +=         jf('<end>')
  p +=     '<instruction_read_second_sharp_sharp>'
  p +=         jf('<instruction_read_third_sharp>')
  p +=     '<instruction_read_third_sharp>'
  p +=     db(instr_sharps)
  p +=     jf('<instruction_read_third_sharp_empty>')
  p +=     jf('<instruction_read_third_sharp_one>')
  p +=     jf('<instruction_read_third_sharp_sharp>')
  p +=     '<instruction_read_third_sharp_empty>'
               ##############################################
               # Opcode is a double sharp (enqueue a sharp) #
               ##############################################
  p +=         dq(instr_ones)
  p +=         eo(enqueue_val_reg)
  p +=         es(enqueue_val_reg)
  p +=         _enqueue_or_dequeue(
                 reg_file       = reg_file,
                 reg_file_tmp   = reg_file_tmp,
                 reg_to_get     = instr_ones,
                 reg_to_get_tmp = instr_ones_tmp,
                 copy_tmp       = copy_tmp,
                 is_enqueue     = True,
                 enqueue_val    = enqueue_val_reg,
                 val_dequeued   = val_dequeued_reg
               )
  p +=         clear(instr_ones) #eo(instr_ones)
  p +=         clear(enqueue_val_reg)
  p +=         eo(instr_idx) # Advance the program counter by one
  p +=         jb('<program_loop>')
               ##############################################
               ##############################################
  p +=     '<instruction_read_third_sharp_one>'
               # Error
  p +=         jf('<end>')
  p +=     '<instruction_read_third_sharp_sharp>'
  p +=         jf('<instruction_read_fourth_sharp>')
  p +=     '<instruction_read_fourth_sharp>'
  p +=     db(instr_sharps)
  p +=     jf('<instruction_read_fourth_sharp_empty>')
  p +=     jf('<instruction_read_fourth_sharp_one>')
  p +=     jf('<instruction_read_fourth_sharp_sharp>')
  p +=     '<instruction_read_fourth_sharp_empty>'
               ###########################################
               # Opcode is a triple sharp (jump forward) #
               ###########################################
  p +=         clear(instr_ones)
  p +=         _jump() # This subprogram will clear out instr_txt (R3)
  p +=         jb('<program_loop>')
               ###########################################
               ###########################################
  p +=     '<instruction_read_fourth_sharp_one>'
               # Error
  p +=         jf('<end>')
  p +=     '<instruction_read_fourth_sharp_sharp>'
  p +=         jf('<instruction_read_fifth_sharp>')
  p +=     '<instruction_read_fifth_sharp>'
  p +=     db(instr_sharps)
  p +=     jf('<instruction_read_fifth_sharp_empty>')
  p +=     jf('<instruction_read_fifth_sharp_one>')
  p +=     jf('<instruction_read_fifth_sharp_sharp>')
  p +=     '<instruction_read_fifth_sharp_empty>'
               ###############################################
               # Opcode is a quadruple sharp (jump backward) #
               ###############################################
  p +=         clear(instr_ones)
  p +=         _jump() # This subprogram will clear out instr_txt (R3)
  p +=         jb('<program_loop>')
               ###############################################
               ###############################################
  p +=     '<instruction_read_fifth_sharp_one>'
               # Error
  p +=         jf('<end>')
  p +=     '<instruction_read_fifth_sharp_sharp>'
  p +=         jf('<instruction_read_hypothetical_sixth_sharp>')
  p +=     '<instruction_read_hypothetical_sixth_sharp>'
  p +=     db(instr_sharps)
  p +=     jf('<instruction_read_hypothetical_sixth_sharp_empty>')
  p +=     jf('<instruction_read_hypothetical_sixth_sharp_one>')
  p +=     jf('<instruction_read_hypothetical_sixth_sharp_sharp>')
  p +=     '<instruction_read_hypothetical_sixth_sharp_empty>'
               #################################################
               # Opcode is pentuple sharp (dequeue and branch) #
               #################################################
  p +=         dq(instr_ones)
  p +=         _enqueue_or_dequeue(
                 reg_file       = reg_file,
                 reg_file_tmp   = reg_file_tmp,
                 reg_to_get     = instr_ones,
                 reg_to_get_tmp = instr_ones_tmp,
                 copy_tmp       = copy_tmp,
                 is_enqueue     = False,
                 enqueue_val    = enqueue_val_reg,
                 val_dequeued   = val_dequeued_reg
               )
  #p +=         eo(instr_ones)     We don't need this anymore because
  #                                instr_ones is not used again for the current
  #                                instruction and it will be repopulated for
  #                                the next instruction when the next
  #                                instruction is executed,
  p +=         clear(instr_ones) # so we clear instr_ones now
  p +=         db(val_dequeued_reg)
  p +=         jf('<db-empty>')
  p +=         jf('<db-one>')
  p +=         jf('<db-sharp>')
  p +=         '<db-empty>'
  p +=             eo(instr_idx)
  p +=             jf('<dequeue_and_branch_return_from>')
  p +=         '<db-one>'
  p +=             eo(instr_idx)
  p +=             eo(instr_idx)
  p +=             jf('<dequeue_and_branch_return_from>')
  p +=         '<db-sharp>'
  p +=             eo(instr_idx)
  p +=             eo(instr_idx)
  p +=             eo(instr_idx)
  p +=             jf('<dequeue_and_branch_return_from>')
  p +=         '<dequeue_and_branch_return_from>'
  p +=         jb('<program_loop>')
               #################################################
               #################################################
  p +=     '<instruction_read_hypothetical_sixth_sharp_one>'
               # Error
  p +=         jf('<end>')
  p +=     '<instruction_read_hypothetical_sixth_sharp_sharp>'
               # Error
  p +=         jf('<end>')
  p += '<epilogue>'
  p +=     clear(instr_idx)
           # Write the encoded value of R1 into the actual program txt reg (R1)
  #p +=     clear(instr_ones) # This is now handled above
  p +=     clear(prog_txt)
  for _ in range(prog_txt-1):
    p +=   eo(instr_ones)
  p +=     '<final-output-loop>'
  p +=       _enqueue_or_dequeue(
               reg_file       = reg_file,
               reg_file_tmp   = reg_file_tmp,
               reg_to_get     = instr_ones,
               reg_to_get_tmp = instr_ones_tmp,
               copy_tmp       = copy_tmp,
               is_enqueue     = False,
               enqueue_val    = enqueue_val_reg,
               val_dequeued   = val_dequeued_reg
             )
  p +=       db(val_dequeued_reg)
  p +=       jf('<check-reg-remaining>')
  p +=       jf('<one>')
  p +=       jf('<sharp>')
  p +=       '<one>'
  p +=         eo(prog_txt)
  p +=         jb('<final-output-loop>')
  p +=       '<sharp>'
  p +=         es(prog_txt)
  p +=         jb('<final-output-loop>')
  p +=       '<check-reg-remaining>'
               # If the reg file encodes only empty registers, we want to clear
               #   it out for a good halt, but if there are any non-empty regs,
               #   we want to leave them so we get a bad halt
  p +=         db(reg_file)
  p +=         jf('<end>')
  p +=         jf('<-one->')
  p +=         jf('<-sharp->')
  p +=         '<-one->'
  #p +=           eo(exception_reg)
  p +=           jf('<end>') # Error
  p +=         '<-sharp->'
  p +=           db(reg_file)
  p +=           jf('<end>') # Error
  p +=           jf('<end>') # Error
  p +=           jb('<check-reg-remaining>')
  p += '<end>'
  return labels_to_offsets(p)



from ..tools.instruction_set import enqueue_one as eo
from ..tools.instruction_set import enqueue_sharp as es
from ..tools.instruction_set import jump_forward as jf
from ..tools.instruction_set import jump_backward as jb
from ..tools.instruction_set import dequeue_and_branch as db
from ..tools.labels_to_offsets import labels_to_offsets
from ..programs.clear import clear
from ..programs.move import move
from ..programs.copy import copy
from ..programs.mult_unary import mult_unary

# pow_unary
# =========
# regA    : positive integer, the index of the register containing the unary
#         :   unary base that will be exponentiated and the register that the
#         :   power will be written to at the end of the computation
# regB    : positive integer, the index of the register containing the unary
#         :   exponent; the register will be returned empty
# tmpAcc  : positive integer, the first temporary register which should be
#         :   passed empty and will be returned empty
# tmpA    : positive integer, the second temporary register which should be
#         :   passed empty and will be returned empty
# tmpAA   : positive integer, the third temporary register which should be
#         :   passed empty and will be returned empty
# tmpBB   : positive integer, the fourth temporary register which should be
#         :   passed empty and will be returned empty
# tmpCopy : positive integer, the fifth temporary register which should be
#         :   passed empty and will be returned empty
# returns : string, the 1# program that when executed will exponentiate the
#         :   contents of regA to the power of regB and write the result back
#         :   into regA
def pow_unary(regA=1, regB=2, tmpAcc=3, tmpA=4, tmpAA=5, tmpBB=6, tmpCopy=7):
  p =  eo(tmpAcc)
  p += '<pow_loop>'
  p +=   db(regB)
  p +=   jf('<empty>')
  p +=   jf('<one>')
  p +=   jf('<sharp>')
  p +=   '<empty>'
  p +=     jf('<epilogue>')
  p +=   '<one>'
  p +=     copy(regA, tmpA, tmpCopy)
  p +=     mult_unary(regA=tmpAcc, regB=tmpA, tmpA=tmpAA, tmpB=tmpBB, tmpCopy=tmpCopy)
  p +=     jb('<pow_loop>')
  p +=   '<sharp>'
  p +=     jf('<error>')
  p += '<epilogue>'
  p +=   clear(regA)
  p +=   clear(regB)
  p +=   move(tmpAcc, regA)
  p += '<error>'
  return labels_to_offsets(p)



from ..tools.instruction_set import enqueue_one as eo
from ..tools.instruction_set import enqueue_sharp as es
from ..tools.instruction_set import jump_forward as jf
from ..tools.instruction_set import jump_backward as jb
from ..tools.instruction_set import dequeue_and_branch as db
from ..tools.labels_to_offsets import labels_to_offsets

# add_unary
# =========
# regA    : positive integer, the index of the register containing the first
#         :   addend and the register in which the sum will be written
# regB    : positive integer, the index of the register containing the second
#         :   addend which will be returned empty
# returns : string, the 1# program that when executed will sum the unary
#         :   numbers contained in regA and regB, writing the result into
#         :   regA and leaving regB empty
def add_unary(regA=1, regB=2):
  p =  '<add_loop>'
  p +=   db(regB)
  p +=   jf('<empty>')
  p +=   jf('<one>')
  p +=   jf('<sharp>')
  p +=   '<empty>'
  p +=     jf('<end>')
  p +=   '<one>'
  p +=     eo(regA)
  p +=     jb('<add_loop>')
  p +=   '<sharp>'
  p +=     jf('<error>')
  p += '<error>'
  p += '<end>'
  return labels_to_offsets(p)
